@@ -1,16 +1,17 @@
 from __future__ import annotations
 
+
 from ._internal_logging import InternalLogger
 import threading
 import time
 from typing import Optional
 
-import sseclient
-import base64
 import prefab_pb2 as Prefab
 import os
-
 from ._count_down_latch import CountDownLatch
+from ._requests import ApiClient, UnauthorizedException
+from ._sse_connection_manager import SSEConnectionManager
+from .config_client_interface import ConfigClientInterface
 from .config_loader import ConfigLoader
 from .config_resolver import ConfigResolver
 from .config_value_unwrapper import ConfigValueUnwrapper
@@ -29,7 +30,7 @@ logger = InternalLogger(__name__)
 class InitializationTimeoutException(Exception):
     def __init__(self, timeout_seconds, key):
         super().__init__(
-            f"Prfeab couldn't initialize in {timeout_seconds} second timeout. Trying to fetch key `{key}`."
+            f"Prefab couldn't initialize in {timeout_seconds} second timeout. Trying to fetch key `{key}`."
         )
 
 
@@ -37,12 +38,11 @@ class MissingDefaultException(Exception):
     def __init__(self, key):
         super().__init__(
             f"""No value found for key '{key}' and no default was provided.
-
 If you'd prefer returning `None` rather than raising when this occurs, modify the `on_no_default` value you provide in your Options."""
         )
 
 
-class ConfigClient:
+class ConfigClient(ConfigClientInterface):
     def __init__(self, base_client):
         self.is_initialized = threading.Event()
         self.checkpointing_thread = None
@@ -50,14 +50,19 @@ class ConfigClient:
         self.sse_client = None
         logger.info("Initializing ConfigClient")
         self.base_client = base_client
-        self.options = base_client.options
+        self._options = base_client.options
         self.init_latch = CountDownLatch()
+        self.unauthorized_event = threading.Event()
         self.finish_init_mutex = threading.Lock()
         self.checkpoint_freq_secs = 60
         self.config_loader = ConfigLoader(base_client)
         self.config_resolver = ConfigResolver(base_client, self.config_loader)
         self._cache_path = None
         self.set_cache_path()
+        self.api_client = ApiClient(self.options)
+        self.sse_connection_manager = SSEConnectionManager(
+            self.api_client, self, self.options.prefab_stream_urls
+        )
 
         if self.options.is_local_only():
             self.finish_init("local only")
@@ -66,7 +71,6 @@ class ConfigClient:
         else:
             # don't load checkpoint here, that'll block the caller. let the thread do it
             self.start_checkpointing_thread()
-            self.start_streaming()
 
     def get(
         self,
@@ -91,6 +95,8 @@ class ConfigClient:
         ok_to_proceed = self.init_latch.wait(
             timeout=self.options.connection_timeout_seconds
         )
+        if self.unauthorized_event.is_set():
+            raise UnauthorizedException(self.options.api_key)
         if not ok_to_proceed:
             if self.options.on_connection_failure == "RAISE":
                 raise InitializationTimeoutException(
@@ -101,6 +107,10 @@ class ConfigClient:
             )
         return self.config_resolver.get(key, context=context)
 
+    @property
+    def options(self):
+        return self._options
+
     def handle_default(self, key, default):
         if default != NoDefaultProvided:
             return default
@@ -109,77 +119,64 @@ class ConfigClient:
         return None
 
     def load_checkpoint(self):
-        if self.load_checkpoint_from_api_cdn():
-            return
-        if self.load_cache():
-            return
-        logger.warning("No success loading checkpoints")
+        try:
+            if self.load_checkpoint_from_api_cdn():
+                return
+            if self.load_cache():
+                return
+            logger.warning("No success loading checkpoints")
+        except UnauthorizedException:
+            self.handle_unauthorized_response()
 
     def start_checkpointing_thread(self):
         self.checkpointing_thread = threading.Thread(
-            target=self.checkpointing_loop, daemon=True
+            target=self.load_checkpoint, daemon=True
         )
         self.checkpointing_thread.start()
 
     def start_streaming(self):
         self.streaming_thread = threading.Thread(
-            target=self.streaming_loop, daemon=True
+            target=self.sse_connection_manager.streaming_loop, daemon=True
         )
         self.streaming_thread.start()
 
-    def streaming_loop(self):
-        while not self.base_client.shutdown_flag.is_set():
-            try:
-                url = "%s/api/v1/sse/config" % self.options.prefab_api_url
-                headers = {
-                    "x-prefab-start-at-id": f"{self.config_loader.highwater_mark}",
-                }
-                response = self.base_client.session.get(
-                    url,
-                    headers=headers,
-                    stream=True,
-                    auth=("authuser", self.options.api_key),
-                    timeout=None,
-                )
+    def is_shutting_down(self):
+        return self.base_client.shutdown_flag.is_set()
 
-                self.sse_client = sseclient.SSEClient(response)
+    def continue_connection_processing(self):
+        return not self.is_shutting_down() and not self.unauthorized_event.is_set()
 
-                for event in self.sse_client.events():
-                    if event.data:
-                        logger.info("Loading data from SSE stream")
-                        configs = Prefab.Configs.FromString(
-                            base64.b64decode(event.data)
-                        )
-                        self.load_configs(configs, "sse_streaming")
-            except Exception as e:
-                if not self.base_client.shutdown_flag.is_set:
-                    logger.info(f"Issue with streaming connection, will restart {e}")
+    def highwater_mark(self) -> int:
+        return self.config_loader.highwater_mark
 
-    def checkpointing_loop(self):
-        while not self.base_client.shutdown_flag.is_set():
-            try:
-                self.load_checkpoint()
-                time.sleep(self.checkpoint_freq_secs)
-            except Exception as e:
-                logger.warn(f"Issue checkpointing, will restart {e}")
+    def load_initial_data(self):
+        try:
+            self.load_checkpoint()
+        except UnauthorizedException:
+            self.handle_unauthorized_response()
 
     def load_checkpoint_from_api_cdn(self):
-        url = "%s/api/v1/configs/0" % self.options.url_for_api_cdn
-        logger.info(f"Loading config from {url}")
-        response = self.base_client.session.get(
-            url, auth=("authuser", self.options.api_key)
-        )
-        if response.ok:
-            configs = Prefab.Configs.FromString(response.content)
-            self.load_configs(configs, "remote_api_cdn")
-            return True
-        else:
-            logger.info(
-                "Checkpoint remote_cdn_api failed to load",
+        try:
+            hwm = self.config_loader.highwater_mark
+            response = self.api_client.resilient_request(
+                "/api/v1/configs/" + str(hwm),
+                auth=("authuser", self.options.api_key),
+                timeout=4,
+                allow_cache=True,
             )
-            return False
+            if response.ok:
+                configs = Prefab.Configs.FromString(response.content)
+                self.load_configs(configs, "remote_api_cdn")
+                return True
+            else:
+                logger.info(
+                    "Checkpoint remote_cdn_api failed to load",
+                )
+                return False
+        except UnauthorizedException:
+            self.handle_unauthorized_response()
 
-    def load_configs(self, configs, source):
+    def load_configs(self, configs: Prefab.Configs, source: str) -> None:
         project_id = configs.config_service_pointer.project_id
         project_env_id = configs.config_service_pointer.project_env_id
         self.config_resolver.project_env_id = project_env_id
@@ -251,6 +248,8 @@ class ConfigClient:
             self.init_latch.count_down()
             if not was_initialized:
                 logger.warning(f"Unlocked config via {source}")
+                if self.options.is_loading_from_api():
+                    self.start_streaming()
                 if self.options.on_ready_callback:
                     threading.Thread(
                         target=self.options.on_ready_callback, daemon=True
@@ -282,6 +281,11 @@ class ConfigClient:
     def is_ready(self) -> bool:
         return self.is_initialized.is_set()
 
+    def handle_unauthorized_response(self):
+        logger.error("Received unauthorized response")
+        self.unauthorized_event.set()
+        self.init_latch.count_down()
+
     def close(self) -> None:
         if self.sse_client:
             self.sse_client.close()
